///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  WriteToFileAsText.cpp                                                             //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//
//<LIBTAG:b6828754-a766-41c2-9fab-582625d641a1>

#include  <fftw3.h>
#include  "modules.h"
#include  "Exception.h"
#include  "{VAL label}.h"


{VAL label}::{VAL label}(void) :
FileOutPath("{VAL params/output_file_path}"),
FileOut(this->FileOutPath,ios::trunc),
CurrSampleNum(0),
FifoIn  ({VAL params/input_fifo}),
FRT(FifoIn)
{
  this->State = State_Initialized;

  if( { VAL debug } )
  { // Debug display Display_Constructor
    cout << this->Display_Constructor();
  }
}


RTIF::ExitStateType  {VAL label}::Init_Allocate(void)
{
  this->State = State_Allocated;

  this->CurrSampleNum = 0;

  if( { VAL debug } )
  { // Display Display_InitReset
    cout << Display_InitAlloc();
  }

  return  RTIF::Exit_Normally;
}


RTIF::ExitStateType  {VAL label}::Init_Reset(void)
{
  this->State = State_Reset;

  // Reset control variables
  this->KillFlag       = false;
  this->Alive          = 0;
  this->Complete       = 0;
  this->CycleCounter   = ~0;
  this->CurrSampleNum  = 0;

  // Spawn the controller thread
  ControlThread = thread( ProcessBatches, ref(*this) );

  if( { VAL debug } )
  { // Display Display_InitReset
    cout << Display_InitReset();
  }

  return  RTIF::Exit_Normally;
}


RTIF::ExitStateType  {VAL label}::ProcessBatches( {VAL label}& _Obj )
{
  // State change
  _Obj.State = State_Processing;

  uint32_t  Samp, K;
  uint32_t  AmountRead = 0U;
  const vector<uint32_t>  LocVect_Out_Incr = { 1, 0 };
  const uint32_t  Fifo_Incr = _Obj.FifoIn.GetBufMap().Offset( LocVect_Out_Incr );
  vector<uint32_t>  LocVect_Out( _Obj.FifoIn.GetVectorLen() );
  {VAL params/data_type}  Sample[_Obj.NumChannels];

  do
  {
    // See if there's any data available
    {
      unique_lock<mutex>   LK(_Obj.GetCvMutex);

      _Obj.CV.wait( LK, [&]{
            return (!WAV_File_In_001_dec.SourceSamplesExhausted) || _Obj.KillFlag; } );

      // Look for a Kill signal
      if( _Obj.KillFlag )  return  RTIF::Exit_Due_To_Kill;
    }

    // FIFO start and lock
    _Obj.FifoIn.FIFO_Read_Any_Amount( _Obj.FRT, AmountRead );

    // Look for a Kill signal
    if( _Obj.KillFlag )  break;

    { // Increment the cycle count to start processing
      unique_lock<mutex>   LK(_Obj.GetCvMutex);
      ++_Obj.CycleCounter;
      LK.unlock();
      _Obj.CV.notify_all();
    }

    // Look for a Kill signal
    if( _Obj.KillFlag )  break;

    /////////////////////////////////
    // Processing occurs here
    /////////////////////////////////

    LocVect_Out[0] = _Obj.FRT.GetReadSchema_View().Start;

    uint32_t  Fifo_Curr = _Obj.FifoIn.GetBufMap().Offset( LocVect_Out );

    for( K=_Obj.CurrSampleNum, Samp=0 ; Samp < AmountRead ; ++K, ++Samp )
    {
      for( uint16_t Chan=0U ; Chan < _Obj.NumChannels ; ++Chan )
      {
        LocVect_Out[1] = Chan;

        // Recover sample
        Sample[Chan] = _Obj.FifoIn.ReadValueFromFifo( Fifo_Curr );

        // Update index into FIFO
        Fifo_Curr = _Obj.FifoIn.IncrFifoPos( Fifo_Curr, Fifo_Incr );
      }

      // Write the data item
      _Obj.FileOut << setw(8) << K << "  [";

      for( uint32_t Chan=0U ; Chan < _Obj.NumChannels ; ++Chan )
      {
        _Obj.FileOut << setw(6) << Sample[Chan];

        if( (Chan+1U) < _Obj.NumChannels )
        {
          _Obj.FileOut << ", ";
        }
      }

      _Obj.FileOut << " ] " << endl;
    }

    // Look for a Kill signal
    if( _Obj.KillFlag )  break;

    // Done writing into FIFO --- alert readers new data is available
    _Obj.FifoIn.FIFO_Read_End(_Obj.FRT);

    // Look for a Kill signal
    if( _Obj.KillFlag )  break;

    // Update the sample counter
    _Obj.CurrSampleNum += AmountRead;

  } while(!_Obj.KillFlag);

  return  RTIF::Exit_Due_To_Kill;
}


RTIF::ExitStateType  {VAL label}::KillThreads(void)
{
  this->State = State_Kill;

  { // Set Kill flags in all FIFOs
    FifoIn.SetKillFlag(true);
  }

  { // Set KillFlag in threads
    unique_lock<mutex>   LK(GetCvMutex);
    this->KillFlag = true;
    LK.unlock();
    CV.notify_all();
  }

  if( { VAL debug } )
  { // Display Display_InitReset
    cout << Display_Kill();
  }

  // Join up with the dying Control Thread
  ControlThread.join();

  return  RTIF::Exit_Due_To_Kill;
}


///////////////////////////////////////
// Display and Debug Routines
///////////////////////////////////////

string  {VAL label}::Display_Constructor(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex);
  ostringstream       oS;

  oS << ">> {VAL label}::Display_Constructor[" << this << "]" << endl;

  // Parameters
  oS << "  -FileOutPath: '" << this->FileOutPath << "'" << endl;
  oS << "  -NumChannels: " << this->NumChannels << endl;
  oS << "  -State: " << DisplayState( this->State ) << endl;

  oS << "<< {VAL label}::Display_Constructor[" << this << "]" << endl << endl;

  return  oS.str();
}


string  {VAL label}::Display_InitAlloc(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex);
  ostringstream       oS;

  oS << ">> {VAL label}::Display_InitAlloc[" << this << "]" << endl;

  oS << "  -NumChannels: " << this->NumChannels << endl;
  oS << "  -Output File Open: " << boolalpha << this->FileOut.is_open() << endl;
  oS << "  -State: " << DisplayState( this->State ) << endl;

  oS << "<< {VAL label}::Display_InitAlloc[" << this << "]" << endl << endl;

  return    oS.str();
}


string  {VAL label}::Display_InitReset(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex);
  ostringstream       oS;

  oS << ">> {VAL label}::Display_InitReset[" << this << "]" << endl;

  oS << "  -NumChannels: " << this->NumChannels << endl;
  oS << "  -Output File Open: " << boolalpha << this->FileOut.is_open() << endl;
  oS << "  -State: " << DisplayState( this->State ) << endl;

  oS << "<< {VAL label}::Display_InitReset[" << this << "]" << endl << endl;

  return    oS.str();
}


string  {VAL label}::Display_Kill(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex);
  ostringstream       oS;

  oS << ">> {VAL label}::Display_Kill[" << this << "]" << endl;

  oS << "  -KillFlag: " << boolalpha << this->KillFlag << endl;
  oS << "  -CycleCounter: " << this->CycleCounter << endl;
  oS << "  -State: " << DisplayState( this->State ) << endl;

  oS << "<< {VAL label}::Display_Kill[" << this << "]" << endl << endl;

  return    oS.str();
}


