///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  BuildBnfcs.cpp                                                                    //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#include  <iostream>
#include  <cassert>
#include  "BuildBnfcs.h"
#include  "IBnfToCppCov.h"
#include  "Exception.h"

using namespace std;


bool  Write_DataDist_PathBuf(void)
{
  const string    FilePath("DataDist/PARSER_Path.bnf0");
  ifstream        GramFile(FilePath);

  if( !GramFile.good() )
  {
    throw  Exception(string("Cannot open input file '" + FilePath + "'."), __FILE__, __LINE__ );
  }

  wej_parser::IBnfToCppCov    Conv(GramFile);

  cout << "[3] Translating Grammar '" << FilePath << "': " << (Conv.IsValidGrammar() ? "<GOOD>" : "<BAD>") << endl;

  if(true)
  {
    ofstream    OutCpp("DataDist/PARSER_Path.cpp",ios::trunc);
    ofstream    OutHpp("DataDist/PARSER_Path.h",ios::trunc);

    OutCpp << Conv.GenerateCppFile( "path_bnf", "PARSER_Path.h");
    OutHpp << Conv.GenerateHppFile( "path_bnf" );
  }

  return  true;
}


bool      Write_XML_PathBuf(void)
{
  const string    FilePath("BNF_XML/PARSER_Xml.bnf0");
  ifstream        GramFile(FilePath);

  if( !GramFile.good() )
  {
    throw  Exception(string("Cannot open input file '" + FilePath + "'."), __FILE__, __LINE__ );
  }

  wej_parser::IBnfToCppCov    Conv(GramFile);

  cout << "[2] Translating Grammar '" << FilePath << "': " << (Conv.IsValidGrammar() ? "<GOOD>" : "<BAD>") << endl;

  if(true)
  {
    ofstream    OutCpp("BNF_XML/XML_Storage_Gram.cpp",ios::trunc);
    ofstream    OutHpp("BNF_XML/XML_Storage_Gram.h",ios::trunc);

    assert(OutCpp.good());
    assert(OutHpp.good());

    OutCpp << Conv.GenerateCppFile( "wej_parser", "XML_Storage_Gram.h" , "wej_xml_parser" );
    OutHpp << Conv.GenerateHppFile( "wej_parser", "wej_xml_parser" );
  }

  return  true;
}


bool      Write_BNFC_PathBuf(void)
{
  const string    FilePath("BNF_XML/PARSER_Bnf0.bnf0");
  ifstream        GramFile(FilePath);

  if( !GramFile.good() )
  {
    throw  Exception(string("Cannot open input file '" + FilePath + "'."), __FILE__, __LINE__ );
  }

  wej_parser::IBnfToCppCov    Conv(GramFile);

  cout << "[1] Translating Grammar '" << FilePath << "': " << (Conv.IsValidGrammar() ? "<GOOD>" : "<BAD>") << endl;

  if(true)
  {
    ofstream    OutCpp("BNF_XML/BNFC_Gen.cpp",ios::trunc);
    ofstream    OutHpp("BNF_XML/BNFC_Gen.h",ios::trunc);

    assert(OutCpp.good());
    assert(OutHpp.good());

    OutCpp << Conv.GenerateCppFile( "wej_parser", "BNFC_Gen.h" , "wej_bnf" );
    OutHpp << Conv.GenerateHppFile( "wej_parser", "wej_bnf" );
  }

  return  true;
}


bool      Write_BNFC_Template(void)
{
  const string    FilePath("DataDist/PARSER_Template.bnf0");
  ifstream        GramFile(FilePath);

  if( !GramFile.good() )
  {
    throw  Exception(string("Cannot open input file '" + FilePath + "'."), __FILE__, __LINE__ );
  }

  wej_parser::IBnfToCppCov    Conv(GramFile);

  cout << "[4] Translating Grammar '" << FilePath << "': " << (Conv.IsValidGrammar() ? "<GOOD>" : "<BAD>") << endl;

  if(true)
  {
    ofstream    OutCpp("DataDist/PARSER_Template.cpp",ios::trunc);
    ofstream    OutHpp("DataDist/PARSER_Template.h",ios::trunc);

    assert(OutCpp.good());
    assert(OutHpp.good());

    OutCpp << Conv.GenerateCppFile( "wej_parser", "PARSER_Template.h" , "wej_template" );
    OutHpp << Conv.GenerateHppFile( "wej_parser", "wej_template" );
  }

  return  true;
}



