///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  CharacterObj.cpp                                                                  //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#include        <cassert>
#include        <list>
#include        <cctype>
#include        <sstream>
#include        <iostream>
#include        <iomanip>
#include        "CharacterObj.h"
#include        "AndRuleObj.h"

using namespace std;


namespace   wej_parser
{
///////////////////////////////////////////////////////////////////////////////////////////////////
// void      CharacterObj::Clear
///////////////////////////////////////////////////////////////////////////////////////////////////

void      CharacterObj::Clear(void)
{
  unsigned int      K;

  for( K=0 ; K < NumNative ; ++K )
  {
    CharArray[K]  = 0;
  }

  // Clear EOF flag
  MatchEOF  = false;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// void      CharacterObj::Invert
///////////////////////////////////////////////////////////////////////////////////////////////////

void      CharacterObj::Invert(void)
{
  unsigned int      K;

  for( K=0 ; K < NumNative ; ++K )
  {
    CharArray[K]  = ~CharArray[K];
  }

  // Leave EOF flag unchanged
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// unsigned int    CharacterObj::CharIntList::Size
///////////////////////////////////////////////////////////////////////////////////////////////////

unsigned int    CharacterObj::CharIntList::Size(void) const
{
  unsigned int                    Num;
  list<CharInt>::const_iterator   I;

  for( I = Letters.begin(), Num = 0 ; I != Letters.end() ; ++I )
  {
    Num   +=  I->Size();
  }

  return    Num;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// string      CharacterObj::Display
///////////////////////////////////////////////////////////////////////////////////////////////////

string      CharacterObj::Display( bool , const unsigned int   ) const
{
  ostringstream     oS;
  bool              InvertedSmaller;

  CharIntList       CL    = GetCharList();
  CharIntList       ICL   = GetInvertedCharList();

  // If the inverted list is shorter display it witha flag
  InvertedSmaller   = (ICL.Size() < CL.Size()) & (ICL.Size() != 0);

  if( !InvertedSmaller )
  {
    oS << "CHAR(";
  }
  else
  {
    CL  = ICL;
    oS << "NCHAR(";
  }

  // Look for special case for CHAR(ANY) or NCHAR(ANY)
  if( (CL.Letters.size() == 1)          &&
      (CL.Letters.front().Low == 0)     &&
      (CL.Letters.front().High == 0xFF)   )
  {
    oS <<"ANY";
  }
  else
  {
    // Display the character intervals
    list<CharInt>::const_iterator     I;

    for( I = CL.Letters.begin() ; I != CL.Letters.end() ; ++I )
    {
      if( I != CL.Letters.begin() )     oS << ',';

      if( I->Low == I->High )
      {
        oS << DisplayCharBnf(I->Low);
      }
      else
      {
        oS << DisplayCharBnf(I->Low) << "-" << DisplayCharBnf(I->High);
      }
    }
  }

  oS << ")";

  oS << DisplayFlags();

  return    oS.str();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// string      CharacterObj::DisplayCharBnf
///////////////////////////////////////////////////////////////////////////////////////////////////

string      CharacterObj::DisplayCharBnf( const GrammarAbs::IntCharType Char )
{
  assert( sizeof(GrammarAbs::CharType) == 1 );      // No Unicode support yet

  ostringstream        oS;

  // Look for EOF
  if( Char == EOF_TOKEN )
  {
    oS << "EOF";
  }
  else if( isgraph(Char) )
  {
    if( Char == ')' )
    {
      oS << "\\)";
    }
    else if( Char == '\\' )
    {
      oS << "\\\\";
    }
    else
    {
      oS << (unsigned char) Char;
    }
  }
  else
  {
    oS << "\\x";

    if( Char < 256 )
    {
      oS << hex << setfill('0') << setw(2) << Char;
    }
    else
    {
      oS << hex << setfill('0') << Char;
    }
  }

  return    oS.str();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// void      CharacterObj::SetChar
///////////////////////////////////////////////////////////////////////////////////////////////////

void      CharacterObj::SetChar( const GrammarAbs::IntCharType Char, bool Value )
{
  if( Char != EOF_TOKEN )
  {
    if(Value)
    {
      CharArray[GetNativeItemNum(Char)]    = CharArray[GetNativeItemNum(Char)] |  GetNativeItemBit(Char);
    }
    else
    {
      CharArray[GetNativeItemNum(Char)]    = CharArray[GetNativeItemNum(Char)] & ~GetNativeItemBit(Char);
    }
  }
  else
  {
    MatchEOF = Value;
  }
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// string      CharacterObj::DisplayCharCpp
///////////////////////////////////////////////////////////////////////////////////////////////////

string      CharacterObj::DisplayCharCpp( const GrammarAbs::IntCharType Char )
{
  assert( sizeof(GrammarAbs::CharType) == 1 );      // No Unicode support yet

  ostringstream        oS;

  // Look for EOF
  if( Char == EOF_TOKEN )
  {
    oS << EOF_TOKEN;
  }
  else if( isgraph(Char) )
  {
    if( Char == '\\' )
    {
      oS << "'\\\\'";
    }
    else if( Char == '\'' )
    {
      oS << "'\\\''";
    }
    else
    {
      oS << "'" << (unsigned char) Char << "'";
    }
  }
  else
  {
    oS << "'" << "\\x";

    if( Char < 256 )
    {
      oS << hex << setfill('0') << setw(2) << Char;
    }
    else
    {
      oS << hex << setfill('0') << Char;
    }

    oS << "'";
  }

  return    oS.str();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// string    CharacterObj::DisplayCharCpp
///////////////////////////////////////////////////////////////////////////////////////////////////

string    CharacterObj::DisplayCharCpp( const string Str )
{
  unsigned int      K;
  string            R, R_Char;
  unsigned char     C;

  for( K=0 ; K < Str.length() ; ++K )
  {
    // Recover the character
    C   = Str[K];

    // See how to represent
    if( isgraph(C) )
    {
      switch(C)
      {
      case  '\'':
        R_Char  = "\\\'";
      break;

      case  '\"':
        R_Char  = "\\\"";
      break;

      case  '\?':
        R_Char  = "\\\?";
      break;

      case  '\\':
        R_Char  = "\\\\";
      break;

      case  '\0':
        R_Char  = "\\0";
      break;

      case  '\a':
        R_Char  = "\\a";
      break;

      case  '\b':
        R_Char  = "\\b";
      break;

      case  '\f':
        R_Char  = "\\f";
      break;

      case  '\n':
        R_Char  = "\\n";
      break;

      case  '\r':
        R_Char  = "\\r";
      break;

      case  '\t':
        R_Char  = "\\t";
      break;

      case  '\v':
        R_Char  = "\\v";
      break;

      default:
        R_Char  = C;
      break;
      }
    }
    else
    {
      ostringstream    oS;
      oS << "\\x" << ios::hex << setw(2) << setfill('0') << C;
      R_Char  = oS.str();
    }

    // Accumulate the output
    R   += R_Char;
  }

  return    R;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// CharacterObj::BitVectorNativeType      CharacterObj::GetChar(
///////////////////////////////////////////////////////////////////////////////////////////////////

CharacterObj::BitVectorNativeType      CharacterObj::GetChar( const GrammarAbs::IntCharType Char ) const
{
  if( Char != EOF_TOKEN )
  { // Non-EOF
    return      CharArray[CharacterObj::GetNativeItemNum(Char)] & CharacterObj::GetNativeItemBit(Char);
  }
  else
  {  // EOF
    return      MatchEOF ? 1 : 0;
  }
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// bool        CharacterObj::Parse
///////////////////////////////////////////////////////////////////////////////////////////////////

bool        CharacterObj::Parse( ParseFrame PF )
{
  char        C;
  bool        R;

  // Set rule
  PF.SetRule(this);

  // Read a character
  if( (PF.ReadCharFromInput(C) == 0) && (PF.IsInEof()) )
  { // Read EOF
    return    MatchEOF;
  }

  // See if it matches
  R    = GetChar(C) != 0;

  if(R)   PF.AcceptAll();

  if(GrammarAbs::DebugParse)
  {
    cout << ">>CHAR  \'" << C << "\'"
         << "  " << (R ? "<++++++>" : "<------>")
         << "  [" << PF.GetRuleDebugDescriptor() << "]"
         << "  [" << PF.GetInPos() << "," << PF.GetOutListPos() << "]"
         << endl;
  }

  return    R;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// void      CharacterObj::SetRangeChars
///////////////////////////////////////////////////////////////////////////////////////////////////

void      CharacterObj::SetRangeChars
          (
            const GrammarAbs::IntCharType     Char0,
            const GrammarAbs::IntCharType     Char1,
            const bool                        Value
          )
{
  GrammarAbs::IntCharType     K;

  assert( Char1 >= Char0 );

  for( K=Char0 ; K <= Char1 ; ++K )
  {
    SetChar( K, Value );
  }
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// string     CharacterObj::GetOutput
///////////////////////////////////////////////////////////////////////////////////////////////////

string     CharacterObj::GetOutput( ParseFrame& PF, const unsigned int RuleNum )
{
  return    PF.GetSegmentChars( RuleNum );
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// CharacterObj::CharIntList     CharacterObj::GetCharList
///////////////////////////////////////////////////////////////////////////////////////////////////

CharacterObj::CharIntList     CharacterObj::GetCharList(void) const
{
  CharIntList     Intervals;

  // Display all the character list
  for( BitVectorNativeType L=0 ; L < ( (1 << this->NativeBits) + 1 ) ; ++L )
  {
    if( !this->GetChar(L) )   continue;
    Intervals.Push(L);
  }

  return    Intervals;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// CharacterObj::CharIntList     CharacterObj::GetInvertedCharList
///////////////////////////////////////////////////////////////////////////////////////////////////

CharacterObj::CharIntList     CharacterObj::GetInvertedCharList(void) const
{
  CharacterObj    CO(*this);
  CO.Invert();
  return      CO.GetCharList();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// void      CharacterObj::GenerateCppParseTree
///////////////////////////////////////////////////////////////////////////////////////////////////

void      CharacterObj::GenerateCppParseTree( IBnfToCppCov& OutputData, bool KleinLayer, string& Includes )
{
  ostringstream     oS;
  UNUSED(Includes);

  oS << "  CHAR = new CharacterObj;" << endl;

  if( KleinLayer )
  {
    oS << "  KLEIN->SetKleinChild(CHAR);" << endl;
  }
  else
  {
    oS << "  AND->PushBack(CHAR);" << endl;
  }

    CharIntList   CL        = GetCharList();
    CharIntList   ICL       = GetInvertedCharList();
    bool          Inverted  = false;

  if( ICL.Letters.size() < CL.Letters.size() )
  {
    CL        = ICL;
    Inverted  = true;
  }

  list<CharInt>::const_iterator     I;

  for( I= CL.Letters.begin() ; I != CL.Letters.end() ; ++I )
  {
    if( I->Low == I->High )
    {
      oS << "  CHAR->SetChar( " << DisplayCharCpp(I->Low) << " );" << endl;
    }
    else
    {
      oS << "  CHAR->SetRangeChars( " << DisplayCharCpp(I->Low) << ", "
                                      << DisplayCharCpp(I->High) << " );" << endl;
    }
  }

  if( Inverted )
  {
    oS << "  CHAR->Invert();" << endl;
  }

  // Accumulate the output
  OutputData.Subroutines.back().Contents    += oS.str();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// string     CharacterObj::CharIntList::DisplayAsInt
///////////////////////////////////////////////////////////////////////////////////////////////////

string     CharacterObj::CharIntList::DisplayAsInt(void) const
{
  list<CharInt>::const_iterator     I;
  ostringstream                     oS;

  oS << "CharMask:[";

  for( I=Letters.begin() ; I != Letters.end() ; ++I )
    {
      if( I != Letters.begin() )  oS << ", ";
      oS << I->DisplayAsInt();
    }

  oS << "]";

  return    oS.str();
}
}   // End namespace 'wej_parser'


