///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  OutputListObj.h                                                                   //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#pragma once

#include    <string>
#include    <list>
#include    "GrammarAbs.h"
#include    "AndRuleObj.h"

using namespace std;


namespace   wej_parser
{
///////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////
// class  OutputListObj : public GrammarAbs
///////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////

class   OutputListObj : public GrammarAbs
{
public:

  class   StatementRef
  {
  private:

    // Parameters
    string          RuleName;
    unsigned int    RuleNum;

  public:

    // If 'pRuleName' is the NULL string use the current rule
    StatementRef( const unsigned int pRuleNum, const string& pRuleName=string() ) :
            RuleName(pRuleName), RuleNum(pRuleNum)
    {
    }

    string    GetOutput( ParseFrame& PF, const unsigned int RuleNum );

    unsigned int  GetRuleNum(void) const
    {
      return   this->RuleNum;
    }

    string  GetRuleName(void) const
    {
      return   this->RuleName;
    }

    string  GetFullNameString(void) const
    {
      ostringstream     oS;

      if( !this->RuleName.empty() )
      {
        oS << this->RuleName << ":" << this->RuleNum;
      }
      else
      {
        oS << this->RuleNum;
      }

      return  oS.str();
    }
  };

public:

  typedef enum eType  { NewEntry, AppendItem, AppendChar }  OutputType;

private:

  // Parameters
  string              OutputLabel;
  list<StatementRef>  RuleNumList;
  OutputType          EnumType;

  // Output cache --- used when one output references another
  OutputRec           CachedOutput;

public:

  inline OutputListObj( OutputType OutputType ) : EnumType(OutputType)
  {
  }

  inline void   SetLabel( const string& Label )
  {
    OutputLabel   = Label;
  }

  // If 'RuleName' is NULL then we use the current rule
  inline void     PushRuleNumBack( const int RuleNum, const string& RuleName=string() )
  {
    RuleNumList.push_back( StatementRef( RuleNum, RuleName ) );
  }

  inline void     SetType( OutputType OutputTypeParam )
  {
    EnumType  = OutputTypeParam;
  }

  inline OutputType     GetOutputType(void) const
  {
    return    EnumType;
  }

  //////////////////////////////
  // Interfaces
  //////////////////////////////

  string            Display( bool SubRule=false, const unsigned int Inden=0 ) const;
  bool              Parse( ParseFrame PF );

  inline string     GetOutput( ParseFrame& PF, const unsigned int RuleNum )
  {
    UNUSED(PF);
    UNUSED(RuleNum);

    // Return the outputs to the output stream
    return    ParseFrame::Display(CachedOutput);
  }

  void    GenerateCppParseTree( IBnfToCppCov& OutputData, bool KleinLayer, string& Includes );

  inline bool   DoesRead(void) const
  {
    return    false;
  }

  bool  InverseParse( InverseParseFrame IPF );
};
}   // End namespace 'wej_parser'


