///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  TagObj.cpp                                                                        //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#include    <iostream>
#include    <iomanip>
#include    "TagObj.h"
#include    "CharacterObj.h"

using namespace std;


namespace  wej_parser
{
///////////////////////////////////////////////////////////////////////////////////////////////////
// string   TagObj::Display
///////////////////////////////////////////////////////////////////////////////////////////////////

string    TagObj::Display( bool , const unsigned int  ) const
{
  ostringstream   oS;
  oS << "TAG(" << this->GetLabel() << ")";
  oS << DisplayFlags();
  return    oS.str();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// bool   TagObj::Parse
///////////////////////////////////////////////////////////////////////////////////////////////////

bool    TagObj::Parse( ParseFrame PF )
{
  // Set rule
  PF.SetRule(this);

  if( PF.IsDebug() )
  {
    cout << ">>Parse TAG[" << this->GetLabel() << ": " << PF.Display() << "]" << endl;
    cout.flush();
  }

  if(GrammarAbs::DebugParse)
  {
    cout << ">>TAG[" << PF.GetInPos() << "," << PF.GetOutListPos() << "]:  \t"
                                                      << this->Display(false) << endl;
  }

  return    true;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// string   TagObj::GetOutput
///////////////////////////////////////////////////////////////////////////////////////////////////

string    TagObj::GetOutput( ParseFrame& PF, const unsigned int RuleNum )
{
  UNUSED(PF);
  UNUSED(RuleNum);

  assert( PF.PosIn >= PF.PosIn0 );

  return  this->GetLabel();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// void   TagObj::GenerateCppParseTree
///////////////////////////////////////////////////////////////////////////////////////////////////

void    TagObj::GenerateCppParseTree( IBnfToCppCov& OutputData, bool KleinLayer, string& Includes )
{
  ostringstream   oS;
  UNUSED(Includes);

  oS << "  LIT = new TagObj;" << endl;

  if( KleinLayer )
  {
    oS << "  KLEIN->SetKleinChild(LIT);" << endl;
  }
  else
  {
    oS << "  AND->PushBack(LIT);" << endl;
  }

  oS << "  LIT->SetValue(\"" << CharacterObj::DisplayCharCpp( this->GetLabel() ) << "\");" << endl;

  // Accumulate the output
  OutputData.Subroutines.back().Contents  += oS.str();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// bool   TagObj::InverseParse
///////////////////////////////////////////////////////////////////////////////////////////////////

bool    TagObj::InverseParse( InverseParseFrame IPF )
{
  // Set rule
  IPF.SetRule(this);

  if( IPF.IsDebug() )
  {
    cout << ">> IParse TAG[" << this->GetLabel() << ":  " << IPF.Display() << "]" << endl;
  }

  // See if we must increment the list position first
  if( !IPF.GetListInInit() )
  {
    // Bump the list to the next TAG
    IPF.IncrementList();
  }
  else
  {
    // Clear the init flag
    IPF.SetListInInit(false);
  }

  // If we are at the end of the list then we return FALSE
  if( IPF.GetListPos() == IPF.GetOutputs().End() )
  {
    if( IPF.IsDebug() )
    {
      cout << "<<-IParse TAG[" << this->GetLabel() << ":  " << IPF.Display() << "]" << endl;
      cout.flush();
    }

    // We ran out of tags
    return  false;
  }

  // Look for TAG match
  auto    R = this->GetLabel() == IPF.GetListPos()->GetDataVect()[0];

  // If we passed
  if(R)
  { // Accept the data on the streams
    IPF.AcceptAll();
  }

  if( IPF.IsDebug() )
  {
    const char    Ret = R ? '+' : '-';

    cout << "<<" << Ret << "IParse TAG[" << this->GetLabel() << ":  " << IPF.Display() << "]" << endl;
  }

  return    R;
}
}   // End namespace 'wej_parser'


