///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  XeqObj.cpp                                                                        //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#include    <iostream>
#include    <iomanip>
#include    <typeinfo>
#include    <map>
#include    "XeqObj.h"
#include    "CharacterObj.h"
#include    "GrammarObj.h"
#include    "AndRuleObj.h"

using namespace std;


namespace   wej_parser
{
///////////////////////////////////////////////////////////////////////////////////////////////////
// string   XeqObj::Display
///////////////////////////////////////////////////////////////////////////////////////////////////

string    XeqObj::Display( bool , const unsigned int  ) const
{
  ostringstream        oS;
  oS << "XEQ(" << this->Link << ")";
  oS << DisplayFlags();
  return    oS.str();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// bool   XeqObj::Parse
///////////////////////////////////////////////////////////////////////////////////////////////////

bool    XeqObj::Parse( ParseFrame PF )
{
  // Set rule
  PF.SetRule(this);

  if( PF.IsDebug() )
  {
    cout << ">>Parse XEQ[" << this->Link << ": " << PF.Display() << "]" << endl;
    cout.flush();
  }

  // This function does nothing in the dynamic scope.
  //    C++ code generated has a call to this subroutine
  if(GrammarAbs::DebugParse)
  {
    cout << "<<XEQ[" << PF.GetInPos() << "," << PF.GetOutListPos() << "]:  "
                                                  << this->Display(false) << "\tR=1" << endl;
  }

  return    true;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// string   XeqObj::GetOutput
///////////////////////////////////////////////////////////////////////////////////////////////////

string    XeqObj::GetOutput( ParseFrame& PF, const unsigned int RuleNum )
{
  UNUSED(PF);
  UNUSED(RuleNum);

  return    string();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// void   XeqObj::GenerateCppParseTree
///////////////////////////////////////////////////////////////////////////////////////////////////

void  XeqObj::GenerateCppParseTree( IBnfToCppCov& OutputData, bool KleinLayer, string& Includes )
{
  ostringstream     oS;
  UNUSED(Includes);

  if( KleinLayer )
  {
    oS << "  KLEIN = new KleinObj;" << endl;
    oS << "  AND->PushBack(KLEIN);" << endl;
    oS << "  KLEIN->PushBack(" << this->Link << ");" << endl;
  }
  else
  {
    oS << "  AND->PushBack(" << this->Link << ");" << endl;
  }

  // Accumulate the output
  OutputData.Subroutines.back().Contents  += oS.str();

  // Add the include for this file
  ostringstream   oI;

  oI << "extern  GrammarAbs*      " << this->Link << ";" << endl;

  Includes  += oI.str();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// bool   XeqObj::InverseParse
///////////////////////////////////////////////////////////////////////////////////////////////////

bool    XeqObj::InverseParse( InverseParseFrame IPF )
{
  // Set rule
  IPF.SetRule(this);

  if( IPF.IsDebug() )
  {
    cout << ">> IParse XEQ[" << this->Link << ":  " << IPF.Display() << "]" << endl;
  }

/*
  // Parse using the first rule
  auto  R   = this->Cache->InverseParse(IPF);

  // If we passed
  if(R)
  { // Accept the data on the streams
    IPF.AcceptAll();
  }
*/

  if( IPF.IsDebug() )
  {
    const bool    R = true;
    const char    Ret = R ? '+' : '-';

    cout << "<<" << Ret << "IParse XEQ[" << this->Link << ":  " << IPF.Display() << "]" << endl;
  }

  return    true;
}
}   // End namespace 'wej_parser'


