///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  FirFilterFourierCoefs_001_lab.cpp                                                 //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//
//<LIBTAG:de887803-6dbf-46a0-9373-29a0053d9adb>

#include  <fftw3.h>
#include  <algorithm>
#include  <complex>
#include  "modules.h"
#include  "Exception.h"
#include  "FirFilterFourierCoefs_001_lab.h"


extern  mutex  FFTW_Mut;


// Definitions of filters
const vector<FirFilterFourierCoefs_001_lab::FilterDef>   FirFilterFourierCoefs_001_lab::FilterDefs
{
  { // Left-Hand Channel
    {
      { LS_FIR_Filter::SegType::eExp_AbsErr, true, 0, 30, 0, 0, 1 },
      { LS_FIR_Filter::SegType::eExp_RelErr, true, 30, 250, 4, 1, 1 },
      { LS_FIR_Filter::SegType::eExp_RelErr, true, 250, 15000, 1, .01, 1 },
      { LS_FIR_Filter::SegType::eExp_AbsErr, true, 15000, 24000, 0, 0, 1 }
    }
  },
  { // Right-Hand Channel
    {
      { LS_FIR_Filter::SegType::eExp_AbsErr, true, 0, 30, 0, 0, 1 },
      { LS_FIR_Filter::SegType::eExp_RelErr, true, 30, 250, 4, 1, 1 },
      { LS_FIR_Filter::SegType::eExp_RelErr, true, 250, 15000, 1, .01, 1 },
      { LS_FIR_Filter::SegType::eExp_AbsErr, true, 15000, 24000, 0, 0, 1 }
    }
  }
};


FirFilterFourierCoefs_001_lab::FirFilterFourierCoefs_001_lab(void) :
NumProcThreads(0),
CoefsFifo_Dimensions  { 1024, 2 },
CoefsFifo_WrapFlags  { true, false },
CoefsFifo_Extent  { CoefsFifo_Dimensions, CoefsFifo_WrapFlags },
CoefsFifo  {CoefsFifo_Extent},
FWT(CoefsFifo),
FilterGenerationComplete {false}
{
  assert( FirFilterFourierCoefs_001_lab::FilterDefs.size() == 2 );

  if( false )
  { // Debug display Display_Constructor
    cout << this->Display_Constructor();
  }
}


RTIF::ExitStateType  FirFilterFourierCoefs_001_lab::Init_Allocate(void)
{
  return  RTIF::Exit_Normally;
}


RTIF::ExitStateType  FirFilterFourierCoefs_001_lab::Init_Reset(void)
{
  // Reset control variables
  this->KillFlag      = false;
  this->Alive         = 0;
  this->Complete      = 0;
  this->CycleCounter  = ~0;
  this->FilterGenerationComplete = false;

  // Calculate 'NumProcThreads'
  if( this->NumProcThreads_FromParams <= this->NumChannels )
  {
    this->NumProcThreads =  this->NumProcThreads_FromParams;
  }
  else
  { // No need to spawn more threads than channels
    this->NumProcThreads = this->NumChannels;
  }

  // We need this value for allocating maximum array sizes during processing
  this->MaxChannelsPerProcThread = 0;

  // Determine which data channels are processed for each thread
  for( uint32_t K=0U ; K < this->NumProcThreads ; ++K )
  {
    BaseChannelNumberForThread( K,
                                this->NumProcThreads,
                                this->NumChannels,
                                this->ChannelStart[K],
                                this->NumChannelsPerThread[K] );

    if( this->NumChannelsPerThread[K] > this->MaxChannelsPerProcThread )
    {
      this->MaxChannelsPerProcThread = this->NumChannelsPerThread[K];
    }
  }

  // Spawn the controller thread
  this->ControlThread = thread( this->ProcessBatches, ref(*this) );

  // Spawn the processing threads
  for( uint32_t K=0U ; K < NumProcThreads ; ++K )
  {
    ThreadCycleCounter[K] = ~0;
    Thread[K]             = thread( ProcessingThread, ref(*this), K );
  }

  { // Wait for all the threads to wake up and then wait on CycleCounter
    unique_lock<mutex>    LK(this->GetCvMutex());
    this->GetCv().wait( LK, [&]{ return (Alive == NumProcThreads) || KillFlag; } );
  }

  if( false )
  { // Display Display_InitReset
    cout << Display_InitReset();
  }

  return  RTIF::Exit_Normally;
}


RTIF::ExitStateType  FirFilterFourierCoefs_001_lab::ProcessBatches( FirFilterFourierCoefs_001_lab& _Obj )
{
  do
  {
    // FIFO start and lock
    _Obj.CoefsFifo.FIFO_Write_Start( _Obj.FirFilterFreqDomainCoefsLen, _Obj.FWT );

    // Look for a Kill signal
    if( _Obj.KillFlag )  return  RTIF::Exit_Due_To_Kill;

    { // Increment the cycle count to start processing
      unique_lock<mutex>   LK(_Obj.GetCvMutex());
      ++_Obj.CycleCounter;
      _Obj.Complete = 0;
      LK.unlock();
      _Obj.GetCv().notify_all();
    }

    // Look for a Kill signal
    if( _Obj.KillFlag )  return  RTIF::Exit_Due_To_Kill;

    /////////////////////////////////
    // Processing occurs here
    /////////////////////////////////

    { // Wait for children to finish
      unique_lock<mutex>   LK(_Obj.GetCvMutex());
      _Obj.GetCv().wait( LK, [&]{ return (_Obj.Complete == _Obj.NumProcThreads) || _Obj.KillFlag; } );
    }

    // Look for a Kill signal
    if( _Obj.KillFlag )  return  RTIF::Exit_Due_To_Kill;

    // Done writing into FIFO --- alert readers new data is available
    _Obj.CoefsFifo.FIFO_Write_End( _Obj.FWT );

    // Look for a Kill signal
    if( _Obj.KillFlag )  return  RTIF::Exit_Due_To_Kill;

    // Indicate that the data is available in the output FIFO now
    _Obj.FilterGenerationComplete = true;

    // Down-stream modules might need this even so new notify_all on our CV
    _Obj.GetCv().notify_all();

  }  while(false);         // Only run once

  return  RTIF::Exit_Normally;
}


RTIF::ExitStateType  FirFilterFourierCoefs_001_lab::ProcessingThread( FirFilterFourierCoefs_001_lab& _Obj, const uint32_t _ThreadNum )
{
  uint32_t  DataChan, RelChannel;
  AVX_Vector<AVX_Vector<complex<double>>>  FilterTimeDomainCoeffcients
                                       (
                                         _Obj.MaxChannelsPerProcThread,
                                         AVX_Vector<complex<double>>(1024)
                                       );
  AVX_Vector<AVX_Vector<complex<double>>>  FilterFreqDomainCoeffcients
                                       (
                                         _Obj.MaxChannelsPerProcThread,
                                         AVX_Vector<complex<double>>(1024)
                                       );
  vector<fftw_plan>         FIRCoefs_TtoF( _Obj.MaxChannelsPerProcThread, nullptr );
  const vector<uint32_t>    LocVect_Out_Incr = { 1, 0 };
  vector<vector<uint32_t>>  LocVect_Out( _Obj.MaxChannelsPerProcThread,
                                vector<uint32_t>(_Obj.CoefsFifo.GetVectorLen()) );

  { // Signal we're alive and ready to process data
    unique_lock<mutex>  LK(_Obj.GetCvMutex());
    ++_Obj.Alive;
    LK.unlock();
    _Obj.GetCv().notify_all();
  }

  // Look for a Kill signal
  if( _Obj.KillFlag )  return  RTIF::Exit_Due_To_Kill;

  // Grab local processing parameters
  _Obj.GetCvMutex().lock();
  const auto    LocChanBase = _Obj.ChannelStart[_ThreadNum];
  const auto    LocNumChans = _Obj.NumChannelsPerThread[_ThreadNum];
  _Obj.GetCvMutex().unlock();

  // Set-up FFT plans
  for( auto K=0U ; K < LocNumChans ; ++K )
  {
    lock_guard<mutex>  Lock(FFTW_Mut);

    if( FIRCoefs_TtoF[K] != nullptr )
    {
      fftw_destroy_plan( FIRCoefs_TtoF[K] );
      FIRCoefs_TtoF[K] = nullptr;
    }

    FIRCoefs_TtoF[K] = fftw_plan_dft_1d
                       (
                         1024,
                         (fftw_complex*) &(FilterTimeDomainCoeffcients[K][0]),
                         (fftw_complex*) &(FilterFreqDomainCoeffcients[K][0]),
                         FFTW_FORWARD,
                         FFTW_ESTIMATE
                       );
  }

  do
  {
    { // Wait for the cycle count to increment
      unique_lock<mutex>   LK(_Obj.GetCvMutex());
      _Obj.GetCv().wait( LK, [&]{ return (_Obj.CycleCounter != _Obj.ThreadCycleCounter[_ThreadNum]) || _Obj.KillFlag; } );
      ++_Obj.ThreadCycleCounter[_ThreadNum];
    }

    // Look for a Kill signal
    if( _Obj.KillFlag )  break;

    /////////////////////////////////
    // Start of Processing
    /////////////////////////////////

    // Set-up and cache the Write Schema
    const auto&  WSchema = _Obj.FWT.GetWriteSchema_View();

    // Set up FIFO acceses for the differnet channels processed here
    for( RelChannel = 0U ; RelChannel < LocNumChans ; ++RelChannel )
    {
      LocVect_Out[RelChannel][0] = WSchema.Start;
      LocVect_Out[RelChannel][1] = LocChanBase + RelChannel;
    }

    // Over data channels
    for( DataChan = LocChanBase ; DataChan < (LocChanBase+LocNumChans) ; ++DataChan )
    {
      RelChannel = DataChan - LocChanBase;

      // Assertions
      assert( 1024 >= _Obj.FirFilterFreqDomainCoefsLen );

      // Add the filter segments
      for( auto I=_Obj.FilterDefs[DataChan].FilterSeg.begin() ;
                          I != _Obj.FilterDefs[DataChan].FilterSeg.end() ; ++I )
      {
        if( I->Symmetric )
        { // Symmetric filter segment
          _Obj.FilterSolutions[DataChan].AddSymmetric
          (
            I->Type,
            I->F0/_Obj.SampleRate,
            I->F1/_Obj.SampleRate,
            I->A0,
            I->A1,
            I->W
          );
        }
        else
        { //  Asymmetric filter segment
          _Obj.FilterSolutions[DataChan].Add
          (
            I->Type,
            I->F0/_Obj.SampleRate,
            I->F1/_Obj.SampleRate,
            I->A0,
            I->A1,
            I->W
          );
        }
      }

      // Generate the filter segments
      _Obj.FilterSolutions[DataChan].GenFilter
               ( _Obj.NumFilterCoefs, FilterTimeDomainCoeffcients[RelChannel] );

      // Post-fill the time domain coefficients with zeros to the length of the FD coefs
      for( auto K=_Obj.NumFilterCoefs ; K < 1024 ; ++K )
      {
        FilterTimeDomainCoeffcients[RelChannel][K] = 0;
      }

      // Perform a Fourier Transform to generate the filters
      fftw_execute_dft( FIRCoefs_TtoF[RelChannel],
                        (fftw_complex*) &(FilterTimeDomainCoeffcients[RelChannel][0]),
                        (fftw_complex*) &(FilterFreqDomainCoeffcients[RelChannel][0])  );

      for( RelChannel=0U ; RelChannel < LocNumChans ; ++RelChannel )
      {
        _Obj.CoefsFifo.WriteVectorToFifo<complex<double>>
        (
          &FilterFreqDomainCoeffcients[RelChannel][0],
          LocVect_Out[RelChannel],
          LocVect_Out_Incr,
          _Obj.FWT.GetWriteSchema_View().Length
        );
      }
    }

    /////////////////////////////////
    // End of Processing
    /////////////////////////////////

    // Show that we are complete and ready to process another batch
    {
      unique_lock<mutex>   LK(_Obj.GetCvMutex());
      ++_Obj.Complete;
      LK.unlock();
      _Obj.GetCv().notify_all();
    }

    // Look for a Kill signal
    if( _Obj.KillFlag )  break;

  }  while( !_Obj.KillFlag );

  // KILL FLAG
  for( auto K=0U ; K < _Obj.MaxChannelsPerProcThread ; ++K )
  {
    if( FIRCoefs_TtoF[K] != nullptr )
    {
      lock_guard<mutex>  Lock(FFTW_Mut);

      fftw_destroy_plan( FIRCoefs_TtoF[K] );
    }
  }

  return  RTIF::Exit_Due_To_Kill;
}


RTIF::ExitStateType  FirFilterFourierCoefs_001_lab::KillThreads(void)
{
  { // Set Kill flags in all FIFOs
    CoefsFifo.SetKillFlag(true);
  }

  { // Set KillFlag in threads
    unique_lock<mutex>   LK(this->GetCvMutex());
    this->KillFlag = true;
    LK.unlock();
    GetCv().notify_all();
  }

  // Join up with the dying Control Thread
  ControlThread.join();

  // Now join up with the dying threads
  for( auto K=0U ; K < this->NumProcThreads ; ++K )
  {
    Thread[K].join();
  }

  if( false )
  {
    cout << Display_Kill();
  }

  return  RTIF::Exit_Due_To_Kill;
}


RTIF::ExitStateType  FirFilterFourierCoefs_001_lab::Exit(void)
{
  { // Set Kill flags in all FIFOs
    CoefsFifo.SetKillFlag(true);
  }

  { // Set KillFlag in threads
    unique_lock<mutex>   LK(this->GetCvMutex());
    this->KillFlag = true;
    LK.unlock();
    GetCv().notify_all();
  }

  // Join up with the dying Control Thread
  ControlThread.join();

  // Now join up with the dying threads
  for( auto K=0U ; K < this->NumProcThreads ; ++K )
  {
    Thread[K].join();
  }

  if( false )
  { // Display Display_InitReset
    cout << Display_Exit();
  }

  return  RTIF::Exit_Normally;
}


///////////////////////////////////////
// Display and Debug Routines
///////////////////////////////////////

string  FirFilterFourierCoefs_001_lab::Display_Constructor(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> FirFilterFourierCoefs_001_lab::Display_Constructor[" << this << "]" << endl;

  // Parameters
  oS << "  -NumProcThreads: " << this->NumProcThreads_FromParams << endl;
  oS << "  -NumChannels: " << this->NumChannels << endl;
  oS << "  -NumFilterCoefs: " << this->NumFilterCoefs << endl;
  oS << "  -FilterGenerationComplete: " << boolalpha << this->FilterGenerationComplete << endl;

  for( const auto& I: this->FilterDefs )
  {
    for( const auto& J: I.FilterSeg )
    {
      oS << "    -SEG:";
      oS << "  F[" << J.F0 << " - " << J.F1 << "]";
      oS << "  Sym[" << boolalpha << J.Symmetric << "]";
      oS << "  A[" << J.A0 << " - " << J.A1 << "]";
      oS << "  W[" << J.W << "]";
      oS << endl;

    }
  }

  oS << "<< FirFilterFourierCoefs_001_lab::Display_Constructor[" << this << "]" << endl << endl;

  return  oS.str();
}


string  FirFilterFourierCoefs_001_lab::Display_InitReset(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> FirFilterFourierCoefs_001_lab::Display_InitReset[" << this << "]" << endl;

  oS << "  -NumProcThreads: " << this->NumProcThreads << endl;
  oS << "  -NumChannels: " << this->NumChannels << endl;
  oS << "  -FilterGenerationComplete: " << boolalpha << this->FilterGenerationComplete << endl;

  for( auto K=0U ; K < 2 ; ++K )
  {
    oS << "    -Thread[" << setw(2) << K << "]:"
       << "  baseChan:" << this->ChannelStart[K]
       << "  numChans:" << this->NumChannelsPerThread[K]
       << endl;
   }

  oS << "<< FirFilterFourierCoefs_001_lab::Display_InitReset[" << this << "]" << endl << endl;

  return    oS.str();
}


string  FirFilterFourierCoefs_001_lab::Display_Kill(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> FirFilterFourierCoefs_001_lab::Display_Kill[" << this << "]" << endl << endl;

  oS << "  -KillFlag: " << boolalpha << this->KillFlag << endl;
  oS << "  -CycleCounter: " << this->CycleCounter << endl;
  oS << "  -FilterGenerationComplete: " << boolalpha << this->FilterGenerationComplete << endl;
  oS << "  -Complete: " << this->Complete << endl;

  for( auto K=0U ; K < 2 ; ++K )
  {
    oS << "    -Thread[" << setw(2) << K << "]:"
       << endl;
   }

  oS << endl << "<< FirFilterFourierCoefs_001_lab::Display_Kill[" << this << "]" << endl << endl;

  return    oS.str();
}


string  FirFilterFourierCoefs_001_lab::Display_Exit(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> FirFilterFourierCoefs_001_lab::Display_Exit[" << this << "]" << endl << endl;

  oS << "  -KillFlag: " << boolalpha << this->KillFlag << endl;
  oS << "  -CycleCounter: " << this->CycleCounter << endl;
  oS << "  -FilterGenerationComplete: " << boolalpha << this->FilterGenerationComplete << endl;
  oS << "  -Complete: " << this->Complete << endl;

  for( auto K=0U ; K < 2 ; ++K )
  {
    oS << "    -Thread[" << setw(2) << K << "]:"
       << endl;
   }

  oS << endl << "<< FirFilterFourierCoefs_001_lab::Display_Exit[" << this << "]" << endl << endl;

  return    oS.str();
}



