///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  WAV_File_In_001_lab.cpp                                                           //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//
//<LIBTAG:7a5fec18-281e-45e1-9051-9f60cc5d1cf3>

#include  <iomanip>
#include  <numbers>
#include  "modules.h"
#include  "WAV_File_In_001_lab.h"


WAV_File_In_001_lab::WAV_File_In_001_lab(void) :
NumProcThreads(0),
InputWavFile { "Test.wav" },
CurrSample {0},
LastSample { InputWavFile.GetNumSamples() },
OutFifo_Dimensions  { 10240, this->NumChannels },
OutFifo_WrapFlags  { true, false },
OutFifo_Extent  { OutFifo_Dimensions, OutFifo_WrapFlags },
OutFifo  {OutFifo_Extent},
FWT(OutFifo),
SourceSamplesExhausted {false}
{
  this->State = State_Initialized;

  if( false )
  { // Debug display Display_Constructor
    cout << this->Display_Constructor();
  }
}


bool  WAV_File_In_001_lab::AnyActiveThreads(void) const
{
  bool  AnyActiveThreads = false;

  AnyActiveThreads = AnyActiveThreads || this->ControlThread.joinable();

  return  AnyActiveThreads;
}


RTIF::ExitStateType  WAV_File_In_001_lab::Init_Allocate(void)
{
  this->State = State_Allocated;

  // If there are any currently active threads kill them and restart
  if( this->AnyActiveThreads() )
  { // If there are active threads: kill them
    this->KillThreads();
    this->KillFlag  = false;
    this->State     = State_Allocated;
  }

  // Reset some control variables
  this->Alive         = 0;
  this->Complete      = 0;
  this->CycleCounter  = ~0;

  if( false )
  { // Display Display_InitAlloc
    cout << Display_InitAlloc();
  }

  return  RTIF::Exit_Normally;
}


RTIF::ExitStateType  WAV_File_In_001_lab::Init_Reset(void)
{
  this->State = State_Reset;

  // If there are any currently active threads kill them and restart
  if( this->AnyActiveThreads() )
  { // There are active threads: kill them
    this->KillThreads();
    this->KillFlag  = false;
    this->State     = State_Allocated;
  }

  // Reset some control variables
  this->CycleCounter            = 0;
  this->CurrSample              = 0;
  this->LastSample              = InputWavFile.GetNumSamples();
  this->SourceSamplesExhausted  = false;

  // Spawn the controller thread
  ControlThread = thread( ProcessBatches, ref(*this) );

  // Look for a Kill signal
  if( this->KillFlag )  return  RTIF::Exit_Due_To_Kill;

  if( false )
  { // Display Display_InitReset
    cout << Display_InitReset();
  }

  return  RTIF::Exit_Normally;
}


RTIF::ExitStateType  WAV_File_In_001_lab::ProcessBatches( WAV_File_In_001_lab& _Obj )
{
  uint32_t                Samp, Chan;
  const vector<uint32_t>  LocVect_Out_Incr = { 1, 0 };
  const uint32_t          Fifo_Incr = _Obj.OutFifo.GetBufMap().Offset( LocVect_Out_Incr );
  vector<uint32_t>        LocVect_Out(_Obj.OutFifo.GetVectorLen());
  vector<uint32_t>        Fifo_Curr(_Obj.NumChannels);

  // Set state
  _Obj.State = State_Processing;

  do
  { // See if we've processed all the data
    assert(_Obj.LastSample >= _Obj.CurrSample);

    { // Wait for some work to show up
      unique_lock<mutex>   LK(_Obj.GetCvMutex());

      _Obj.GetCv().wait( LK, [&]{ return (_Obj.LastSample >= _Obj.CurrSample) ||
                                          _Obj.KillFlag; } );
    }

    // Look for a Kill signal
    if( _Obj.KillFlag )  return  RTIF::Exit_Due_To_Kill;

    // Look for end of data
    if( _Obj.LastSample == _Obj.CurrSample )  break;

    // Determine how much data we can write into the FIFO
    uint32_t  NumFifoSlotsToReserve = 2048;

    if( NumFifoSlotsToReserve > (_Obj.LastSample - _Obj.CurrSample) )
    {
      NumFifoSlotsToReserve = (_Obj.LastSample - _Obj.CurrSample);
    }

    // FIFO start and lock
    _Obj.OutFifo.FIFO_Write_Start( NumFifoSlotsToReserve, _Obj.FWT );

    // Look for a Kill signal
    if( _Obj.KillFlag )  return  RTIF::Exit_Due_To_Kill;

    // Set-up and cache the Write Schema
    const auto&  WSchema = _Obj.FWT.GetWriteSchema_View();

    ///////////////////////////////////
    // Copy the samples into the FIFO
    ///////////////////////////////////

    // Vector Elements:  [ Sample Num, Channel Num ]
    // Fifo Elements:    [ FIFO.PROC, Channel Num ]

    LocVect_Out[0] = WSchema.Start;

    for( Chan=0U ; Chan < _Obj.NumChannels ; ++Chan )
    {
      LocVect_Out[1]  = Chan;
      Fifo_Curr[Chan] = _Obj.OutFifo.GetBufMap().Offset( LocVect_Out );
    }

    for( Samp=0U ; Samp < NumFifoSlotsToReserve ; ++Samp )
    for( Chan=0U ; Chan < _Obj.NumChannels ; ++Chan )
    {
      _Obj.OutFifo.WriteValueToFifo
      (
        _Obj.InputWavFile.GetSample_s16( Samp+_Obj.CurrSample, Chan ),
        Fifo_Curr[Chan]
      );

      Fifo_Curr[Chan] = _Obj.OutFifo.IncrFifoPos( Fifo_Curr[Chan], Fifo_Incr );
    }

    /////////////////////////////////
    // End of processing
    /////////////////////////////////

    // Done writing into FIFO --- alert readers new data is available
    _Obj.OutFifo.FIFO_Write_End( _Obj.FWT );

    // Look for a Kill signal
    if( _Obj.KillFlag )  return  RTIF::Exit_Due_To_Kill;

    // These are the number of audio samples processed
    _Obj.CurrSample += NumFifoSlotsToReserve;

    // Update counters
    ++_Obj.CycleCounter;

  } while( !_Obj.KillFlag && (_Obj.CurrSample < _Obj.LastSample) );

  // Set EndOfData flag in output FIFO so downstream readers will be alerted
  if( _Obj.LastSample == _Obj.CurrSample )
  {
    _Obj.OutFifo.SetEndOfDataFlag();

    // We need to wait till all our consumers recover the data they need.
    //   We will be issued a Kill signal eventually.
    { // Wait for some work to show up
      unique_lock<mutex>   LK(_Obj.GetCvMutex());

      _Obj.GetCv().wait( LK, [&]{ return _Obj.KillFlag.load() || _Obj.OutFifo.IsFifoComplete() ; } );
    }
  }

  if( false )
  { // Debug display
    cout << _Obj.Display_ProcessBatches();
  }

  // Look for a Kill signal
  if( _Obj.KillFlag )  return  RTIF::Exit_Due_To_Kill;

  return  RTIF::Exit_Normally;
}


RTIF::ExitStateType  WAV_File_In_001_lab::KillThreads(void)
{
  this->State = State_Kill;

  { // Set Kill flags in all FIFOs
    OutFifo.SetKillFlag(true);
  }

  { // Set KillFlag in threads
    unique_lock<mutex>   LK(this->GetCvMutex());
    this->KillFlag = true;
    LK.unlock();
    this->GetCv().notify_all();
  }

  // Join up with the dying Control Thread
  ControlThread.join();

  if( false )
  { // Debug display
    cout << this->Display_Kill();
  }

  return  RTIF::Exit_Due_To_Kill;
}


RTIF::ExitStateType  WAV_File_In_001_lab::Exit(void)
{
  this->State = State_Exit;

  { // Set Kill flags in all FIFOs
    OutFifo.SetKillFlag(true);
  }

  { // Set KillFlag in threads
    unique_lock<mutex>   LK(this->GetCvMutex());
    this->KillFlag = true;
    LK.unlock();
    this->GetCv().notify_all();
  }

  // Join up with the dying Control Thread
  ControlThread.join();

  if( false )
  { // Debug display
    cout << this->Display_Exit();
  }

  return  RTIF::Exit_Normally;
}


///////////////////////////////////////
// Display and Debug Routines
///////////////////////////////////////

string  WAV_File_In_001_lab::Display_Constructor(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> WAV_File_In_001_lab::Constructor[" << this << "]" << endl;

  // Parameters
  oS << "  -NumChannels: " << this->NumChannels << endl;
  oS << "  -SourceWavFilePath: '" << this->Source_WAV_FilePath << "'" << endl;
  oS << "  -BigEndianCPU: " << boolalpha << WaveFileIn::IsThisA_BigEndianMachine() << endl;
  oS << "  -State: " << DisplayState( this->State ) << endl;

  oS << InputWavFile.Display_Header(2);

  oS << "<< WAV_File_In_001_lab::Constructor[" << this << "]" << endl << endl;

  return  oS.str();
}


string  WAV_File_In_001_lab::Display_InitAlloc(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> WAV_File_In_001_lab::InitAlloc[" << this << "]" << endl;

  oS << "  -NumChannels: " << this->NumChannels << endl;
  oS << "  -WAV File Open: " << boolalpha << this->InputWavFile.IsOpen() << endl;
  oS << "  -State: " << DisplayState( this->State ) << endl;

  oS << "<< WAV_File_In_001_lab::InitAlloc[" << this << "]" << endl << endl;

  return    oS.str();
}


string  WAV_File_In_001_lab::Display_InitReset(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> WAV_File_In_001_lab::InitReset[" << this << "]" << endl;

  oS << "  -NumChannels: " << this->NumChannels << endl;
  oS << "  -WAV File Open: " << boolalpha << this->InputWavFile.IsOpen() << endl;
  oS << "  -State: " << DisplayState( this->State ) << endl;

  oS << "<< WAV_File_In_001_lab::InitReset[" << this << "]" << endl << endl;

  return    oS.str();
}


string  WAV_File_In_001_lab::Display_Kill(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> WAV_File_In_001_lab::Kill[" << this << "]" << endl;

  oS << "  -KillFlag: " << boolalpha << this->KillFlag << endl;
  oS << "  -CycleCounter: " << this->CycleCounter << endl;
  oS << "  -State: " << DisplayState( this->State ) << endl;

  oS << "<< WAV_File_In_001_lab::Kill[" << this << "]" << endl << endl;

  return    oS.str();
}


string  WAV_File_In_001_lab::Display_Exit(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> WAV_File_In_001_lab::Exit[" << this << "]" << endl;

  oS << "  -KillFlag: " << boolalpha << this->KillFlag << endl;
  oS << "  -CycleCounter: " << this->CycleCounter << endl;
  oS << "  -State: " << DisplayState( this->State ) << endl;

  oS << "<< WAV_File_In_001_lab::Exit[" << this << "]" << endl << endl;

  return    oS.str();
}


string  WAV_File_In_001_lab::Display_ProcessBatches(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> WAV_File_In_001_lab::ProcessBatches[" << this << "]" << endl;

  oS << "  -KillFlag: " << boolalpha << this->KillFlag << endl;
  oS << "  -CycleCounter: " << this->CycleCounter << endl;
  oS << "  -State: " << DisplayState( this->State ) << endl;
  oS << "  -Sample: " << this->CurrSample << " / " << this->LastSample << endl;

  oS << "<< WAV_File_In_001_lab::ProcessBatches[" << this << "]" << endl << endl;

  return    oS.str();
}


