///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  WAV_File_Out_001_lab.cpp                                                          //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//
//<LIBTAG:3b549cd4-8d58-4bcc-b8b2-9a443918b174>

#include  <iomanip>
#include  "modules.h"
#include  "WAV_File_Out_001_lab.h"


WAV_File_Out_001_lab::WAV_File_Out_001_lab(void) :
NumProcThreads(0),
OutputWavFile { "Test_Out_1.wav",
                 2,
                 16,
                 48000 },
OutputFilePath { "Test_Out_1.wav" },
CurrSampleNum(0),
FifoIn(OverlapAndAddFiltering_001_lab_dec.OutFifo),
FRT_In(FifoIn)
{
  this->State = State_Initialized;

  if( false )
  { // Debug display Display_Constructor
    cout << this->Display_Constructor();
  }
}


bool  WAV_File_Out_001_lab::AnyActiveThreads(void) const
{
  bool  AnyActiveThreads = false;

  AnyActiveThreads = AnyActiveThreads || this->ControlThread.joinable();

  return  AnyActiveThreads;
}


RTIF::ExitStateType  WAV_File_Out_001_lab::Init_Allocate(void)
{
  this->State = State_Allocated;

  // If there are any currently active threads kill them and restart
  if( this->AnyActiveThreads() )
  { // There are active threads: kill them
    this->KillThreads();
    this->KillFlag  = false;
    this->State     = State_Allocated;
  }

  // Reset control variables
  this->CycleCounter  = 0;
  this->CurrSampleNum = 0;

  if( false )
  { // Display Display_InitAlloc
    cout << Display_InitAlloc();
  }

  return  RTIF::Exit_Normally;
}


RTIF::ExitStateType  WAV_File_Out_001_lab::Init_Reset(void)
{
  this->State = State_Reset;

  // If there are any currently active threads kill them and restart
  if( this->AnyActiveThreads() )
  { // There are active threads: kill them
    this->KillThreads();
    this->KillFlag  = false;
    this->State     = State_Reset;
  }

  // Reset control variables
  this->CycleCounter  = 0;
  this->CurrSampleNum = 0;

  // Spawn the controller thread
  ControlThread = thread( ProcessBatches, ref(*this) );

  if( false )
  { // Display Display_InitReset
    cout << Display_InitReset();
  }

  return  RTIF::Exit_Normally;
}


RTIF::ExitStateType  WAV_File_Out_001_lab::ProcessBatches( WAV_File_Out_001_lab& _Obj )
{
  _Obj.State = State_Processing;

  uint32_t    Samp = 0U;
  uint32_t    Chan = 0U;
  uint32_t    NumFifoSamplesToRead = 0U;
  const vector<uint32_t>  LocVect_In_Incr = { 1, 0 };
  const uint32_t          Fifo_Incr = _Obj.FifoIn.GetBufMap().Offset( LocVect_In_Incr );
  vector<uint32_t>        LocVect_In(_Obj.FifoIn.GetVectorLen());
  vector<uint32_t>        Fifo_Curr(_Obj.NumChannels);

  // The Processing Loop
  do
  {
    // See if there is any more data to process
    if( _Obj.FifoIn.IsFifoComplete() )  break;

    _Obj.FifoIn.FIFO_Read
    (
      _Obj.FifoIn.READ_ANY_EXCEPT_NO_DATA,
      NumFifoSamplesToRead,
      _Obj.FRT_In
    );

    // Look for a Kill signal
    if( _Obj.KillFlag )  return  RTIF::Exit_Due_To_Kill;

    // Set-up and cache the Write Schema
    const auto&  InSchema = _Obj.FRT_In.GetReadSchema_View();

    LocVect_In[0] = InSchema.Start;

    for( Chan=0U ; Chan < _Obj.NumChannels ; ++Chan )
    {
      LocVect_In[1]   = Chan;
      Fifo_Curr[Chan] = _Obj.FifoIn.GetBufMap().Offset( LocVect_In );
    }

    for( Samp=0U ; Samp < _Obj.FRT_In.GetActualReadLength() ; ++Samp )
    for( Chan=0U ; Chan < _Obj.NumChannels ; ++Chan )
    {
      _Obj.OutputWavFile.SetSample_s16
      (
        Samp + _Obj.CurrSampleNum,
        Chan,
        _Obj.FifoIn.ReadValueFromFifo( Fifo_Curr[Chan] )
      );

      Fifo_Curr[Chan] = _Obj.FifoIn.IncrFifoPos( Fifo_Curr[Chan], Fifo_Incr );
    }

    // Look for a Kill signal
    if( _Obj.KillFlag )  return  RTIF::Exit_Due_To_Kill;

    ///////////////////////////
    // Data is available here
    ///////////////////////////

    _Obj.FifoIn.FIFO_Read_End( _Obj.FRT_In );

    // Samples just accumulated
    _Obj.CurrSampleNum += _Obj.FRT_In.GetActualReadLength();

    ++_Obj.CycleCounter;

  }  while( !_Obj.KillFlag && !_Obj.FifoIn.IsFifoComplete() );

  // Make sure the header reflects the length of the WAV File
  _Obj.OutputWavFile.CloseFileHeader();

  if( false )
  { // Debug display
    cout << _Obj.Display_ProcessBatches();
  }

  // Look for a Kill signal
  if( _Obj.KillFlag )  return  RTIF::Exit_Due_To_Kill;

  return  RTIF::Exit_Normally;
}


RTIF::ExitStateType  WAV_File_Out_001_lab::KillThreads(void)
{
  this->State = State_Kill;

  { // Set Kill flags in all FIFOs
    this->FifoIn.SetKillFlag(true);
  }

  { // Set KillFlag in threads
    unique_lock<mutex>   LK(this->GetCvMutex());
    this->KillFlag = true;
    LK.unlock();
    GetCv().notify_all();
  }

  // Join up with the dying Control Thread
  ControlThread.join();

  if( false )
  { // Debug display
    cout << this->Display_Kill();
  }

  return  RTIF::Exit_Due_To_Kill;
}


RTIF::ExitStateType  WAV_File_Out_001_lab::Exit(void)
{
  this->State = State_Exit;

  { // Set Kill flags in all FIFOs
    this->FifoIn.SetKillFlag(true);
  }

  { // Set KillFlag in threads
    unique_lock<mutex>   LK(this->GetCvMutex());
    this->KillFlag = true;
    LK.unlock();
    GetCv().notify_all();
  }

  // Join up with the dying Control Thread
  ControlThread.join();

  if( false )
  { // Debug display
    cout << this->Display_Kill();
  }

  return  RTIF::Exit_Normally;
}


///////////////////////////////////////
// Display and Debug Routines
///////////////////////////////////////

string  WAV_File_Out_001_lab::Display_Constructor(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> WAV_File_Out_001_lab::Constructor[" << this << "]" << endl;

  // Parameters
  oS << "  -NumChannels: " << this->NumChannels << endl;
  oS << "  -OutputWavFilePath: '" << this->OutputFilePath << "'" << endl;
  oS << "  -BigEndianCPU: " << boolalpha << WaveFileIn::IsThisA_BigEndianMachine() << endl;
  oS << "  -State: " << DisplayState( this->State ) << endl;

  oS << OutputWavFile.Display_Header(2);

  oS << "<< WAV_File_Out_001_lab::Constructor[" << this << "]" << endl << endl;

  return  oS.str();
}


string  WAV_File_Out_001_lab::Display_InitReset(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> WAV_File_Out_001_lab::InitReset[" << this << "]" << endl;

  oS << "  -NumChannels: " << this->NumChannels << endl;
  oS << "  -WAV File Open: " << boolalpha << this->OutputWavFile.IsFileOpen() << endl;
  oS << "  -State: " << DisplayState( this->State ) << endl;

  oS << "<< WAV_File_Out_001_lab::InitReset[" << this << "]" << endl << endl;

  return    oS.str();
}


string  WAV_File_Out_001_lab::Display_InitAlloc(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> WAV_File_Out_001_lab::InitAlloc[" << this << "]" << endl;

  oS << "  -NumChannels: " << this->NumChannels << endl;
  oS << "  -WAV File Open: " << boolalpha << this->OutputWavFile.IsFileOpen() << endl;
  oS << "  -State: " << DisplayState( this->State ) << endl;

  oS << "<< WAV_File_Out_001_lab::InitAlloc[" << this << "]" << endl << endl;

  return    oS.str();
}


string  WAV_File_Out_001_lab::Display_Kill(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> WAV_File_Out_001_lab::Kill[" << this << "]" << endl;

  oS << "  -KillFlag: " << boolalpha << this->KillFlag << endl;
  oS << "  -CycleCounter: " << this->CycleCounter << endl;
  oS << "  -State: " << DisplayState( this->State ) << endl;

  oS << "<< WAV_File_Out_001_lab::Kill[" << this << "]" << endl << endl;

  return    oS.str();
}


string  WAV_File_Out_001_lab::Display_Exit(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> WAV_File_Out_001_lab::Exit[" << this << "]" << endl;

  oS << "  -KillFlag: " << boolalpha << this->KillFlag << endl;
  oS << "  -CycleCounter: " << this->CycleCounter << endl;
  oS << "  -State: " << DisplayState( this->State ) << endl;

  oS << "<< WAV_File_Out_001_lab::Exit[" << this << "]" << endl << endl;

  return    oS.str();
}


string  WAV_File_Out_001_lab::Display_ProcessBatches(void)
{
  lock_guard<mutex>   LK(this->GetCvMutex());
  ostringstream       oS;

  oS << ">> WAV_File_Out_001_lab::ProcessBatches[" << this << "]" << endl;

  oS << "  -KillFlag: " << boolalpha << this->KillFlag << endl;
  oS << "  -CycleCounter: " << this->CycleCounter << endl;
  oS << "  -State: " << DisplayState( this->State ) << endl;
  oS << "  -Sample: " << this->CurrSampleNum << " / " << WAV_File_In_001_lab_dec.LastSample << endl;

  oS << "<< WAV_File_Out_001_lab::ProcessBatches[" << this << "]" << endl << endl;

  return    oS.str();
}


