///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  Exception.cpp                                                                     //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#include "Exception.h"


Exception::Exception( const string&       _FileName,
                      const unsigned int  _LineNumber ) :
FileName(_FileName),
LineNumber(_LineNumber)
{
}


Exception::Exception( const string&       _Message,
                      const string&       _FileName,
                      const unsigned int  _LineNumber ) :
FileName(_FileName),
LineNumber(_LineNumber)
{
  Message << _Message;
}


Exception::Exception( const Exception&  _Other ) :
FileName(_Other.FileName),
LineNumber(_Other.LineNumber)
{
  Message << _Other.Message.str();
}


string  Exception::Display(void) const
{
  ostringstream   oS;

  // Strip the file name so as to remove the directory string
  size_t    PosOfLastSlash = this->FileName.find_last_of("/");

  // Display the exception
  if( PosOfLastSlash == string::npos )
  { // No directory at beginning of file name
    oS << "*Exception[" << FileName << ":" << LineNumber << "] " << Message.str();
  }
  else
  {
    oS << "*Exception[" << FileName.substr(PosOfLastSlash+1)
       << ":" << LineNumber << "] " << Message.str();
  }

  return  oS.str();
}


