///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  OrRuleObj.cpp                                                                     //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#include  <cassert>
#include  <iostream>
#include  <iomanip>
#include  <sstream>
#include  <typeinfo>
#include  <sstream>
#include  "OrRuleObj.h"
#include  "CharacterObj.h"
#include  "BNFC_Gen.h"
#include  "KleinWrapperObj.h"
#include  "RefObj.h"
#include  "LiteralObj.h"
#include  "OutputListObj.h"
#include  "AndRuleObj.h"
#include  "GrammarObj.h"

using namespace std;


namespace   wej_parser
{
///////////////////////////////////////////////////////////////////////////////////////////////////
// OrRuleObj::SpaceBetweenRules
///////////////////////////////////////////////////////////////////////////////////////////////////

const char*   OrRuleObj::SpaceBetweenRules = "  ";


///////////////////////////////////////////////////////////////////////////////////////////////////
// OrRuleObj::~OrRuleObj
///////////////////////////////////////////////////////////////////////////////////////////////////

OrRuleObj::~OrRuleObj()
{
  unsigned int    K;

  for( K=0 ; K < Group.size() ; ++K )
  {
    delete  Group[K];
  }
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// string   OrRuleObj::Display
///////////////////////////////////////////////////////////////////////////////////////////////////

string      OrRuleObj::Display( bool SubRule, const unsigned int Indent ) const
{
  ostringstream     oS;
  bool              First;
  unsigned int      K, L;

  for( K=0 ; K < Indent ; ++K )   oS << ' ';
  oS  << this->GetLabel();
  oS  << ":" << endl;

  if(SubRule)
  {
    oS << SpaceBetweenRules << "(";
  }

  for( L=0, First=true ; L < Group.size() ; ++L, First=false )
  {
    if( !First )
    {
      for( K=0 ; K < Indent ; ++K )   oS << ' ';
      oS << "|" << endl;
    }

    for( K=0 ; K < (Indent+2) ; ++K )   oS << ' ';
    oS << Group[L]->Display(false,Indent) << endl;
  }

  if(SubRule)
  {
    oS << ")" << SpaceBetweenRules;
  }
  else
  {
    for( K=0 ; K<Indent ; ++K )   oS << ' ';
    oS << ";" << endl;
  }

  oS << DisplayFlags();

  return    oS.str();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// bool   OrRuleObj::Parse
///////////////////////////////////////////////////////////////////////////////////////////////////

bool    OrRuleObj::Parse( ParseFrame PF )
{
  unsigned int    K;
  bool            R;

  // Set rule
  PF.SetRule(this);

  if( PF.IsDebug() )
  {
    cout << ">>Parse OR[" << this->GetLabel() << ": " << PF.Display() << "]" << endl;
  }

  for( K=0, R=true ; K < Group.size() ; ++K )
  {
    ParseFrame      PF1(PF);

    R = Group[K]->Parse(PF);

    if(R)
    {
      // Accept the data on the streams
      PF1.AcceptAll();
      PF.AcceptAll();

      break;       // This worked
    }
  }

  if(GrammarAbs::DebugParse)
  {
    cout << ">>OR "
         << "  " << (R ? "<++++++>" : "<------>")
         << "  [" << PF.GetRuleDebugDescriptor() << "]"
         << "  [" << PF.GetInPos() << "," << PF.GetOutListPos() << "]"
         << endl;
  }

  return    R;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// string   OrRuleObj::GetOutput
///////////////////////////////////////////////////////////////////////////////////////////////////

string  OrRuleObj::GetOutput( ParseFrame& PF, const unsigned int RuleNum )
{
  UNUSED(PF);
  UNUSED(RuleNum);

  return  string();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// void      OrRuleObj::GenerateCppParseTree
///////////////////////////////////////////////////////////////////////////////////////////////////

void      OrRuleObj::GenerateCppParseTree( IBnfToCppCov& OutputData, bool KleinLayer, string& Includes )
{
  unsigned int    K;
  AndRuleObj*     AND;
  ostringstream   oS0;

  for( K=0 ; K < Group.size() ; ++K )
  {
    ostringstream       oS;

    AND = dynamic_cast<AndRuleObj*>(Group[K]);

    oS  << endl;
    oS  << "  ///////////////////////////////////////////////////////////" << endl;
    oS  << "  // AND Group #" << K << endl;
    oS  << "  ///////////////////////////////////////////////////////////" << endl;
    oS  << endl;
    oS  << "  AND = new AndRuleObj;" << endl;
    oS  << "  OR->PushBack(AND);" << endl;

    // Accumulate the output
    OutputData.Subroutines.back().Contents  += oS.str();

    // Get lower-level constructs
    AND->GenerateCppParseTree( OutputData, KleinLayer, Includes );
  }
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// bool   OrRuleObj::InverseParse
///////////////////////////////////////////////////////////////////////////////////////////////////

bool    OrRuleObj::InverseParse( InverseParseFrame IPF )
{
  // Set rule
  IPF.SetRule(this);

  assert(!Group.empty());

  // If Debug
  if( IPF.IsDebug() )
  {
    cout << ">> IParse OR[" << this->GetLabel() << ": " << IPF.Display() << "]" << endl;
  }

  // Over the rules in the group
  for( auto S: Group )
  {
    // Parse using the first rule
    auto  R   = S->InverseParse(IPF);

    // If we passed
    if(R)
    { // Accept the data on the streams
      IPF.AcceptAll();

      if( IPF.IsDebug() )
      {
        cout << "<<+IParse OR[" << IPF.Display() << "]" << endl;
      }

      return  true;
    }
  }

  if( IPF.IsDebug() )
  {
    cout << "<<-IParse OR[" << IPF.Display() << "]" << endl;
  }

  return  false;
}
}   // End namespace 'wej_parser'


