///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  PARSER_Path.cpp                                                                   //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#include        <cassert>
#include        <iostream>
#include        <iomanip>
#include        <sstream>
#include        <typeinfo>
#include        "GrammarAbs.h"
#include        "GrammarObj.h"
#include        "CharacterObj.h"
#include        "BNFC_Gen.h"
#include        "KleinWrapperObj.h"
#include        "RefObj.h"
#include        "LiteralObj.h"
#include        "OutputListObj.h"
#include        "AndRuleObj.h"
#include        "OrRuleObj.h"
#include        "PARSER_Path.h"

using namespace std;
using namespace wej_parser;


namespace path_bnf
{
// Variables used in subroutines
GrammarObj*       GRAM;
CharacterObj*     CHAR;
KleinWrapperObj*  KLEIN;
RefObj*           REF;
LiteralObj*       LIT;
OutputListObj*    OUTL;
AndRuleObj*       AND;
OrRuleObj*        OR;

// Function Prototypes
GrammarAbs*    Parse_Start(void);
GrammarAbs*    Parse_DoubleSlash(void);
GrammarAbs*    Parse_Slash(void);
GrammarAbs*    Parse_Label(void);
GrammarAbs*    Parse_SlashLabel(void);
GrammarAbs*    Parse_Symbol(void);
GrammarAbs*    Parse_LabelSymbol(void);
GrammarAbs*    Parse_WhitespaceSeqOpt(void);
GrammarAbs*    Parse_WhitespaceSeq(void);
GrammarAbs*    Parse_Whitespace(void);

// XEQ Includes


GrammarAbs*   Parse_Path(void)
{
  ///////////////////////////////////////////////////////////
  // GRAM Group
  ///////////////////////////////////////////////////////////

  GRAM  = new GrammarObj;
  GRAM->SetLabel("Path");

  ///////////////////////////////////////////////////////////
  // Rule Entry Points
  ///////////////////////////////////////////////////////////

  GRAM->PushBack(Parse_Start());
  GRAM->PushBack(Parse_DoubleSlash());
  GRAM->PushBack(Parse_Slash());
  GRAM->PushBack(Parse_Label());
  GRAM->PushBack(Parse_SlashLabel());
  GRAM->PushBack(Parse_Symbol());
  GRAM->PushBack(Parse_LabelSymbol());
  GRAM->PushBack(Parse_WhitespaceSeqOpt());
  GRAM->PushBack(Parse_WhitespaceSeq());
  GRAM->PushBack(Parse_Whitespace());

  return  GRAM;
}


GrammarAbs*    Parse_Start(void)
{
  ///////////////////////////////////////////////////////////
  // OR Group 'Start'
  ///////////////////////////////////////////////////////////

  OR  = new OrRuleObj;
  OR->SetLabel("Start");

  ///////////////////////////////////////////////////////////
  // AND Group #0
  ///////////////////////////////////////////////////////////

  AND = new AndRuleObj;
  OR->PushBack(AND);

  // RULE #1: 'REF(DoubleSlash)'
  REF = new RefObj;
  REF->SetLink("DoubleSlash");
  AND->PushBack(REF);

  // RULE #2: 'REF(WhitespaceSeqOpt)'
  REF = new RefObj;
  REF->SetLink("WhitespaceSeqOpt");
  AND->PushBack(REF);

  // RULE #3: 'CHAR(EOF)'
  CHAR = new CharacterObj;
  AND->PushBack(CHAR);
  CHAR->SetChar( 256 );

  ///////////////////////////////////////////////////////////
  // AND Group #1
  ///////////////////////////////////////////////////////////

  AND = new AndRuleObj;
  OR->PushBack(AND);

  // RULE #1: 'REF(DoubleSlash)'
  REF = new RefObj;
  REF->SetLink("DoubleSlash");
  AND->PushBack(REF);

  // RULE #2: 'REF(Label)'
  REF = new RefObj;
  REF->SetLink("Label");
  AND->PushBack(REF);

  // RULE #3: 'REF(SlashLabel)*'
  KLEIN    = new KleinWrapperObj;
  KLEIN->SetKlein(GrammarAbs::KlienStar);
  AND->PushBack(KLEIN);
  REF = new RefObj;
  REF->SetLink("SlashLabel");
  KLEIN->SetKleinChild(REF);

  // RULE #4: 'REF(WhitespaceSeqOpt)'
  REF = new RefObj;
  REF->SetLink("WhitespaceSeqOpt");
  AND->PushBack(REF);

  // RULE #5: 'CHAR(EOF)'
  CHAR = new CharacterObj;
  AND->PushBack(CHAR);
  CHAR->SetChar( 256 );

  ///////////////////////////////////////////////////////////
  // AND Group #2
  ///////////////////////////////////////////////////////////

  AND = new AndRuleObj;
  OR->PushBack(AND);

  // RULE #1: 'REF(Slash)'
  REF = new RefObj;
  REF->SetLink("Slash");
  AND->PushBack(REF);

  // RULE #2: 'REF(WhitespaceSeqOpt)'
  REF = new RefObj;
  REF->SetLink("WhitespaceSeqOpt");
  AND->PushBack(REF);

  // RULE #3: 'CHAR(EOF)'
  CHAR = new CharacterObj;
  AND->PushBack(CHAR);
  CHAR->SetChar( 256 );

  ///////////////////////////////////////////////////////////
  // AND Group #3
  ///////////////////////////////////////////////////////////

  AND = new AndRuleObj;
  OR->PushBack(AND);

  // RULE #1: 'REF(Slash)'
  REF = new RefObj;
  REF->SetLink("Slash");
  AND->PushBack(REF);

  // RULE #2: 'REF(Label)'
  REF = new RefObj;
  REF->SetLink("Label");
  AND->PushBack(REF);

  // RULE #3: 'REF(SlashLabel)*'
  KLEIN    = new KleinWrapperObj;
  KLEIN->SetKlein(GrammarAbs::KlienStar);
  AND->PushBack(KLEIN);
  REF = new RefObj;
  REF->SetLink("SlashLabel");
  KLEIN->SetKleinChild(REF);

  // RULE #4: 'REF(WhitespaceSeqOpt)'
  REF = new RefObj;
  REF->SetLink("WhitespaceSeqOpt");
  AND->PushBack(REF);

  // RULE #5: 'CHAR(EOF)'
  CHAR = new CharacterObj;
  AND->PushBack(CHAR);
  CHAR->SetChar( 256 );

  ///////////////////////////////////////////////////////////
  // AND Group #4
  ///////////////////////////////////////////////////////////

  AND = new AndRuleObj;
  OR->PushBack(AND);

  // RULE #1: 'REF(Label)'
  REF = new RefObj;
  REF->SetLink("Label");
  AND->PushBack(REF);

  // RULE #2: 'REF(SlashLabel)*'
  KLEIN    = new KleinWrapperObj;
  KLEIN->SetKlein(GrammarAbs::KlienStar);
  AND->PushBack(KLEIN);
  REF = new RefObj;
  REF->SetLink("SlashLabel");
  KLEIN->SetKleinChild(REF);

  // RULE #3: 'REF(WhitespaceSeqOpt)'
  REF = new RefObj;
  REF->SetLink("WhitespaceSeqOpt");
  AND->PushBack(REF);

  // RULE #4: 'CHAR(EOF)'
  CHAR = new CharacterObj;
  AND->PushBack(CHAR);
  CHAR->SetChar( 256 );

  return    OR;
}


GrammarAbs*    Parse_DoubleSlash(void)
{
  ///////////////////////////////////////////////////////////
  // OR Group 'DoubleSlash'
  ///////////////////////////////////////////////////////////

  OR  = new OrRuleObj;
  OR->SetLabel("DoubleSlash");

  ///////////////////////////////////////////////////////////
  // AND Group #0
  ///////////////////////////////////////////////////////////

  AND = new AndRuleObj;
  OR->PushBack(AND);

  // RULE #1: 'REF(WhitespaceSeqOpt)'
  REF = new RefObj;
  REF->SetLink("WhitespaceSeqOpt");
  AND->PushBack(REF);

  // RULE #2: 'CHAR(/)'
  CHAR = new CharacterObj;
  AND->PushBack(CHAR);
  CHAR->SetChar( '/' );

  // RULE #3: 'CHAR(/)'
  CHAR = new CharacterObj;
  AND->PushBack(CHAR);
  CHAR->SetChar( '/' );

  // RULE #4: '$(D//)'
  OUTL = new OutputListObj(OutputListObj::NewEntry);
  AND->PushBack(OUTL);
  OUTL->SetLabel("D//");

  return    OR;
}


GrammarAbs*    Parse_Slash(void)
{
  ///////////////////////////////////////////////////////////
  // OR Group 'Slash'
  ///////////////////////////////////////////////////////////

  OR  = new OrRuleObj;
  OR->SetLabel("Slash");

  ///////////////////////////////////////////////////////////
  // AND Group #0
  ///////////////////////////////////////////////////////////

  AND = new AndRuleObj;
  OR->PushBack(AND);

  // RULE #1: 'REF(WhitespaceSeqOpt)'
  REF = new RefObj;
  REF->SetLink("WhitespaceSeqOpt");
  AND->PushBack(REF);

  // RULE #2: 'CHAR(/)'
  CHAR = new CharacterObj;
  AND->PushBack(CHAR);
  CHAR->SetChar( '/' );

  // RULE #3: '$(/)'
  OUTL = new OutputListObj(OutputListObj::NewEntry);
  AND->PushBack(OUTL);
  OUTL->SetLabel("/");

  return    OR;
}


GrammarAbs*    Parse_Label(void)
{
  ///////////////////////////////////////////////////////////
  // OR Group 'Label'
  ///////////////////////////////////////////////////////////

  OR  = new OrRuleObj;
  OR->SetLabel("Label");

  ///////////////////////////////////////////////////////////
  // AND Group #0
  ///////////////////////////////////////////////////////////

  AND = new AndRuleObj;
  OR->PushBack(AND);

  // RULE #1: 'REF(WhitespaceSeqOpt)'
  REF = new RefObj;
  REF->SetLink("WhitespaceSeqOpt");
  AND->PushBack(REF);

  // RULE #2: 'REF(Symbol)'
  REF = new RefObj;
  REF->SetLink("Symbol");
  AND->PushBack(REF);

  return    OR;
}


GrammarAbs*    Parse_SlashLabel(void)
{
  ///////////////////////////////////////////////////////////
  // OR Group 'SlashLabel'
  ///////////////////////////////////////////////////////////

  OR  = new OrRuleObj;
  OR->SetLabel("SlashLabel");

  ///////////////////////////////////////////////////////////
  // AND Group #0
  ///////////////////////////////////////////////////////////

  AND = new AndRuleObj;
  OR->PushBack(AND);

  // RULE #1: 'REF(WhitespaceSeqOpt)'
  REF = new RefObj;
  REF->SetLink("WhitespaceSeqOpt");
  AND->PushBack(REF);

  // RULE #2: 'CHAR(/)'
  CHAR = new CharacterObj;
  AND->PushBack(CHAR);
  CHAR->SetChar( '/' );

  // RULE #3: 'REF(WhitespaceSeqOpt)'
  REF = new RefObj;
  REF->SetLink("WhitespaceSeqOpt");
  AND->PushBack(REF);

  // RULE #4: 'REF(Symbol)'
  REF = new RefObj;
  REF->SetLink("Symbol");
  AND->PushBack(REF);

  return    OR;
}


GrammarAbs*    Parse_Symbol(void)
{
  ///////////////////////////////////////////////////////////
  // OR Group 'Symbol'
  ///////////////////////////////////////////////////////////

  OR  = new OrRuleObj;
  OR->SetLabel("Symbol");

  ///////////////////////////////////////////////////////////
  // AND Group #0
  ///////////////////////////////////////////////////////////

  AND = new AndRuleObj;
  OR->PushBack(AND);

  // RULE #1: 'REF(LabelSymbol)'
  REF = new RefObj;
  REF->SetLink("LabelSymbol");
  AND->PushBack(REF);

  // RULE #2: '$(LAB,1)'
  OUTL = new OutputListObj(OutputListObj::NewEntry);
  AND->PushBack(OUTL);
  OUTL->SetLabel("LAB");
  OUTL->PushRuleNumBack(1);

  return    OR;
}


GrammarAbs*    Parse_LabelSymbol(void)
{
  ///////////////////////////////////////////////////////////
  // OR Group 'LabelSymbol'
  ///////////////////////////////////////////////////////////

  OR  = new OrRuleObj;
  OR->SetLabel("LabelSymbol");

  ///////////////////////////////////////////////////////////
  // AND Group #0
  ///////////////////////////////////////////////////////////

  AND = new AndRuleObj;
  OR->PushBack(AND);

  // RULE #1: 'CHAR(A-Z,_,a-z)'
  CHAR = new CharacterObj;
  AND->PushBack(CHAR);
  CHAR->SetRangeChars( 'A', 'Z' );
  CHAR->SetChar( '_' );
  CHAR->SetRangeChars( 'a', 'z' );

  // RULE #2: 'CHAR(--.,0-9,A-Z,_,a-z)*'
  KLEIN    = new KleinWrapperObj;
  KLEIN->SetKlein(GrammarAbs::KlienStar);
  AND->PushBack(KLEIN);
  CHAR = new CharacterObj;
  KLEIN->SetKleinChild(CHAR);
  CHAR->SetRangeChars( '-', '.' );
  CHAR->SetRangeChars( '0', '9' );
  CHAR->SetRangeChars( 'A', 'Z' );
  CHAR->SetChar( '_' );
  CHAR->SetRangeChars( 'a', 'z' );

  ///////////////////////////////////////////////////////////
  // AND Group #1
  ///////////////////////////////////////////////////////////

  AND = new AndRuleObj;
  OR->PushBack(AND);

  // RULE #1: 'CHAR(.)'
  CHAR = new CharacterObj;
  AND->PushBack(CHAR);
  CHAR->SetChar( '.' );

  // RULE #2: 'CHAR(.)'
  CHAR = new CharacterObj;
  AND->PushBack(CHAR);
  CHAR->SetChar( '.' );

  return    OR;
}


GrammarAbs*    Parse_WhitespaceSeqOpt(void)
{
  ///////////////////////////////////////////////////////////
  // OR Group 'WhitespaceSeqOpt'
  ///////////////////////////////////////////////////////////

  OR  = new OrRuleObj;
  OR->SetLabel("WhitespaceSeqOpt");

  ///////////////////////////////////////////////////////////
  // AND Group #0
  ///////////////////////////////////////////////////////////

  AND = new AndRuleObj;
  OR->PushBack(AND);

  // RULE #1: 'CHAR(\x09-\x0a,\x0d,\x20)*'
  KLEIN    = new KleinWrapperObj;
  KLEIN->SetKlein(GrammarAbs::KlienStar);
  AND->PushBack(KLEIN);
  CHAR = new CharacterObj;
  KLEIN->SetKleinChild(CHAR);
  CHAR->SetRangeChars( '\x09', '\x0a' );
  CHAR->SetChar( '\x0d' );
  CHAR->SetChar( '\x20' );

  return    OR;
}


GrammarAbs*    Parse_WhitespaceSeq(void)
{
  ///////////////////////////////////////////////////////////
  // OR Group 'WhitespaceSeq'
  ///////////////////////////////////////////////////////////

  OR  = new OrRuleObj;
  OR->SetLabel("WhitespaceSeq");

  ///////////////////////////////////////////////////////////
  // AND Group #0
  ///////////////////////////////////////////////////////////

  AND = new AndRuleObj;
  OR->PushBack(AND);

  // RULE #1: 'CHAR(\x09-\x0a,\x0d,\x20)+'
  KLEIN    = new KleinWrapperObj;
  KLEIN->SetKlein(GrammarAbs::KlienPlus);
  AND->PushBack(KLEIN);
  CHAR = new CharacterObj;
  KLEIN->SetKleinChild(CHAR);
  CHAR->SetRangeChars( '\x09', '\x0a' );
  CHAR->SetChar( '\x0d' );
  CHAR->SetChar( '\x20' );

  return    OR;
}


GrammarAbs*    Parse_Whitespace(void)
{
  ///////////////////////////////////////////////////////////
  // OR Group 'Whitespace'
  ///////////////////////////////////////////////////////////

  OR  = new OrRuleObj;
  OR->SetLabel("Whitespace");

  ///////////////////////////////////////////////////////////
  // AND Group #0
  ///////////////////////////////////////////////////////////

  AND = new AndRuleObj;
  OR->PushBack(AND);

  // RULE #1: 'CHAR(\x09-\x0a,\x0d,\x20)'
  CHAR = new CharacterObj;
  AND->PushBack(CHAR);
  CHAR->SetRangeChars( '\x09', '\x0a' );
  CHAR->SetChar( '\x0d' );
  CHAR->SetChar( '\x20' );

  return    OR;
}



}

