///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  xml_decorator.cpp                                                                 //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#include    <cstdlib>
#include    <iostream>
#include    <iomanip>
#include    <fstream>
#include    <filesystem>
#include    <chrono>
#include    <malloc.h>
#include    "DataPool.h"
#include    "DataPath.h"
#include    "DataRef.h"
#include    "DataNodeRef.h"
#include    "PARSER_Template.h"
#include    "BuildBnfcs.h"
#include    "TemplateRenderer.h"
#include    "DataNodeRef.h"
#include    "GenerateFiles.h"

using namespace std;
namespace fs = std::filesystem;


extern  string PES_Tools_Root;

///////////////////////////////////////////////////////////////////////////////////////////////////
// End GlobalSystemFilePath
///////////////////////////////////////////////////////////////////////////////////////////////////

wej_parser::GrammarAbs*   TemplateParser = nullptr;

///////////////////////////////////////////////////////////////////////////////////////////////////
// int  main
///////////////////////////////////////////////////////////////////////////////////////////////////
//
//  Call Parameters:
//
//    TemplateRenderer  <Data.xml path>  <XmlExecRoot>  <OutputPath>
//    TemplateRenderer  <Data.xml path>  <OutputPathOpt>
//
///////////////////////////////////////////////////////////////////////////////////////////////////

int  main( int NumArgs, char** ArgList )
{
  try
  {
  const bool  Debug = false;
  const bool  RebuildBnfTranslators = false;

  // For time-tagging output files
  chrono::time_point<chrono::system_clock>  Now = chrono::system_clock::now();
  auto  InTimeT = chrono::system_clock::to_time_t(Now);
  ostringstream   oTimeDate;
  oTimeDate << put_time( localtime(&InTimeT), "%Y%m%d_%H%M%S" );
  const string  CurrTimeString = oTimeDate.str();

  //////////////////////
  // Handle parameters
  //////////////////////

    fs::path  ProgramXml = fs::current_path();
    fs::path  BuildPath;

    switch(NumArgs)
    {
      case 0:
      case 1:
        cout << "[!]Error: Call as \'xml_decorator  <PROGRAM.XML>  {<OUTPUT_PATH=current>}\'[!]" << endl;

        return	1;
      break;

      case 2:
        ProgramXml = ArgList[1];
        BuildPath  = fs::current_path();

        cout << ">>xml_decorator:  Source:" << ProgramXml << "  Target:" << BuildPath << endl;
     break;

      default:
        ProgramXml = ArgList[1];
        BuildPath  = ArgList[2];

        cout << ">>xml_decorator:  Source:" << ProgramXml << "  Target:" << BuildPath << endl;
      break;
    }

    if(Debug)
    {
      cout << endl;
      cout << "+[exe] xml_decorator" << endl;
      cout << "-[param 'ProgramXml'] " << ProgramXml << endl;
      cout << "-[param 'BuildPath']  " << BuildPath << endl;
    }

    // Update the grammars [OFFLINE MODE]
    if(RebuildBnfTranslators)
    { // Translate: BNF ---> CPP
      cout << endl << ">>BNF0 ---> C++ Grammar Translation Stage Starting" << endl;

      Write_BNFC_PathBuf();
      Write_XML_PathBuf();
      Write_DataDist_PathBuf();
      Write_BNFC_Template();

      cout << "<<BNF0 ---> C++ Grammar Translation Stage Complete" << endl;

      return  2;
    }

    ////////////////////////////////
    // Paths
    ////////////////////////////////

    const fs::path  LibManifestXml = fs::absolute(fs::path(PES_Tools_Root) /
                                                  fs::path("lib_manifest.xml"));
    ProgramXml = fs::absolute(ProgramXml);
    BuildPath = fs::absolute(BuildPath);

    if(Debug)
    {
      cout << "-LibManifestXml: " << LibManifestXml.lexically_normal() << endl;
      cout << "-ProgramXml:     " << ProgramXml.lexically_normal() << endl;
      cout << "-BuildPath:      " << BuildPath.lexically_normal() << endl;
      cout.flush();
    }

    ////////////////////////////////
    // PROGRAM.XML
    ////////////////////////////////

    wej_data_pool::DataTrees.AddSource
    (
      new DataTree
      (
        ProgramXml,
        DataTree::eMODE_LOAD_FROM_FILE,
        DataPath("//Program")
      )
    );

    if(Debug)
    {
      cout << "<PROGRAM.XML> " << ProgramXml << endl;
      cout.flush();
    }

    ////////////////////////////////
    // LIB_MANIFEST.XML
    ////////////////////////////////

    wej_data_pool::DataTrees.AddSource
    (
      new DataTree
      (
        LibManifestXml,
        DataTree::eMODE_LOAD_FROM_FILE,
        DataPath("//Library")
      )
    );

    if(Debug)
    {
      cout << "<LIB_MANIFEST.XML> " << LibManifestXml << endl;
      cout.flush();
    }

    // WEJTAG
    if(Debug)
    {
      cout << endl;
      cout << wej_data_pool::DataTrees.Display();
    }

    //////////////////////////////////////////////////////
    // Clear the target directory before writing into it
    //////////////////////////////////////////////////////

    // Delete the output directory tree
    filesystem::remove_all( BuildPath );

    // Re-create the directory
    const bool  Ret = filesystem::create_directory( BuildPath );

    if(!Ret)
    {
      ostringstream   oS;
      oS << "Cannot create rendered template directory '" << BuildPath << "'";
      throw  Exception(oS.str(),__FILE__,__LINE__);
    }

    if(false)
    {
      cout << "<Decorate> " << BuildPath << endl;
      cout.flush();
    }

    ////////////////////////////////////
    // Process the Template Definition
    ////////////////////////////////////

    // Iterate over the modules to create
    DataNodeRef  ModulesChild( DataPath("//Program/Modules"), true );

    // Render all the templates
    for( ModulesChild.Begin() ; ModulesChild.MoreChildren() ; ModulesChild.Next() )
    {
      ProcessTemplates( ModulesChild.Curr(), BuildPath, CurrTimeString );
    }

    ///////////////////////////////////////////////////////////////////////////
    // Reorder the modules so referenced modules come before the referencee's
    ///////////////////////////////////////////////////////////////////////////

    // Reorder the modules so referenced modules come before the referencee's
    RefRec::ReorderList(Modules);

    ////////////////////////////////////////////////
    // Create the synthesized processors main file
    ////////////////////////////////////////////////

    Init_StartCPP_File( BuildPath, CurrTimeString );
    Init_ModulesCPP_File( BuildPath, CurrTimeString );
    Init_ModulesH_File( BuildPath, CurrTimeString );

    ////////////////////////////////////////////////
    // Say goodbye
    ////////////////////////////////////////////////

    cout << "<<xml_decorator:  Complete at " << CurrTimeString << endl;
  }

  catch(Exception&  DE)
  {
    cout << endl << DE.Display() << endl << endl;
    return  1;
  }

  return    0;
}


